/*
 * Calendar manager object
 *
 * This keeps track of a given calendar.  Eventually this will abtract everything
 * related to getting calendars/saving calendars locally or to a remote Calendar Service
 *
 * Copyright (C) 1998, 1999 the Free Software Foundation
 *
 * Authors:
 *   Miguel de Icaza (miguel@gnu.org)
 *   Federico Mena (quartic@gimp.org)
 *
 * Changed in 2002 by Sebastian Rittau <srittau@jroger.in-berlin.de>
 * for the GNOME 2 port.
 */

#include <config.h>

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <gnome.h>

#include <mimedir/mimedir-datetime.h>
#include <mimedir/mimedir-valarm.h>
#include <mimedir/mimedir-vcal.h>
#include <mimedir/mimedir-vcomponent.h>
#include <mimedir/mimedir-vevent.h>
#include <mimedir/mimedir-vtodo.h>

#include "calendar.h"
#include "gnomecal-alarm-sink.h"
#include "gnomecal-ui.h"
#include "timeutil.h"

#ifdef HAVE_TZNAME
extern char *tzname[2];
#endif

/* FIXME: */
#define MIMEDIR_IS_VJOURNAL(o) 0

/* Our day range */
time_t calendar_day_begin, calendar_day_end;

void calendar_set_day (void);

/*
 * Generate every possible event.  Invokes the callback routine for
 * every occurrence of the event in the [START, END] time interval.
 *
 * If END is zero, the event is generated forever.
 * The callback routine is expected to return 0 when no further event
 * generation is requested.
 */
static void
ical_object_generate_events (MIMEDirVComponent *comp, time_t start, time_t end, calendarfn cb, void *closure)
{
	GList *occurences, *list;
	MIMEDirDateTime *start_dt, *end_dt;

	start_dt = mimedir_datetime_new_from_time_t (start);
	end_dt   = mimedir_datetime_new_from_time_t (end);

	occurences = mimedir_vcomponent_get_occurences_between (comp, start_dt, end_dt);

	for (list = occurences; list != NULL; list = g_list_next (list)) {
		MIMEDirDateTime *date_start, *date_end = NULL, *st, *e;

		g_assert (list->data != NULL && MIMEDIR_IS_DATETIME (list->data));
		date_start = MIMEDIR_DATETIME (list->data);

		list = g_list_next (list);
		g_assert (list != NULL);
		g_assert (list->data == NULL || MIMEDIR_IS_DATETIME (list->data));
		if (list->data)
			date_end = MIMEDIR_DATETIME (list->data);

		if (mimedir_datetime_compare (start_dt, date_start) < 0)
			st = date_start;
		else
			st = start_dt;
		e = end_dt;
		if (date_end && mimedir_datetime_compare (end_dt, date_end) > 0)
				e = date_end;

		(* cb) (comp, mimedir_datetime_get_time_t (st), mimedir_datetime_get_time_t (e), closure);
	}

	mimedir_vcomponent_free_occurences (occurences);

	g_object_unref (G_OBJECT (start_dt));
	g_object_unref (G_OBJECT (end_dt));
}

Calendar *
calendar_new (void)
{
	Calendar *cal;

	cal = g_new0 (Calendar, 1);

	if ((calendar_day_begin == 0) || (calendar_day_end == 0))
		calendar_set_day ();

	cal->event_hash = g_hash_table_new (g_str_hash, g_str_equal);

	cal->alarm_sink = gncal_alarm_sink_new ();
	
	return cal;
}


void
calendar_add_object (Calendar *cal, MIMEDirVComponent *comp)
{
	gchar *uid;

	g_return_if_fail (cal != NULL);
	g_return_if_fail (comp != NULL);
	g_return_if_fail (MIMEDIR_IS_VCOMPONENT (comp));
	
	g_object_get (G_OBJECT (comp), "uid", &uid, NULL);

	if (MIMEDIR_IS_VEVENT (comp)) {
		g_object_ref (G_OBJECT (comp));
		g_hash_table_insert (cal->event_hash, uid, comp);
		cal->events = g_list_prepend (cal->events, comp);

		gncal_alarm_sink_add (cal->alarm_sink, comp);
	}

	else if (MIMEDIR_IS_VTODO (comp)) {
		/* dummy */
	}

	else if (MIMEDIR_IS_VJOURNAL (comp)) {
		g_object_ref (G_OBJECT (comp));
		cal->journal = g_list_prepend (cal->journal, comp);

		gncal_alarm_sink_add (cal->alarm_sink, comp);
	}

	else
		g_return_if_reached ();

	if (!uid) {
		uid = g_strdup_printf ("GnomeCalendar-%ld\n", time (NULL));
		g_object_set (G_OBJECT (comp), "uid", uid, NULL);
	}

	g_free (uid);
	
	cal->modified = TRUE;
}

void
calendar_remove_object (Calendar *cal, MIMEDirVComponent *comp)
{
	if (MIMEDIR_IS_VEVENT (comp)) {
		cal->events = g_list_remove (cal->events, comp);
		g_hash_table_remove (cal->event_hash, comp);
		g_object_unref (G_OBJECT (comp));

		gncal_alarm_sink_remove (cal->alarm_sink, comp);
	}

	else if (MIMEDIR_IS_VTODO (comp)) {
		/* dummy */
	}

	else if (MIMEDIR_IS_VJOURNAL (comp)) {
		cal->journal = g_list_remove (cal->journal, comp);
		g_object_unref (G_OBJECT (comp));

		gncal_alarm_sink_remove (cal->alarm_sink, comp);
	}

	else
		g_assert_not_reached ();

	cal->modified = TRUE;
}

void
calendar_free (Calendar *cal)
{
	GList *item;

	for (item = cal->events; item != NULL; item = g_list_next (item))
		g_object_unref (G_OBJECT (item->data));
	g_list_free (cal->events);
	cal->events = NULL;

	for (item = cal->journal; item != NULL; item = g_list_next (item))
		g_object_unref (G_OBJECT (item->data));
	g_list_free (cal->journal);
	cal->journal = NULL;

	g_hash_table_destroy (cal->event_hash);
	cal->event_hash = g_hash_table_new (g_str_hash, g_str_equal);

	gncal_alarm_sink_clear (cal->alarm_sink);
}

void
calendar_destroy (Calendar *cal)
{
	g_hash_table_destroy (cal->event_hash);
	calendar_free (cal);

	g_object_unref (G_OBJECT (cal->alarm_sink));
	
	g_free (cal);
}

void
calendar_iterate (Calendar *cal, time_t start, time_t end, calendarfn cb, void *closure)
{
	GList *list;

	for (list = cal->events; list != NULL; list = g_list_next (list)) {
		MIMEDirVComponent *comp;

		g_assert (MIMEDIR_IS_VCOMPONENT (list->data));
		comp = MIMEDIR_VCOMPONENT (list->data);

		ical_object_generate_events (comp, start, end, cb, closure);
	}
}

void
calendar_set_day (void)
{
	time_t calendar_today;
	
	calendar_today     = time (NULL);
	calendar_day_begin = time_day_begin (calendar_today);
	calendar_day_end   = time_day_end (calendar_today);
}

static gint
calendar_object_compare_by_start (gconstpointer a, gconstpointer b)
{
	const CalendarObject *ca = a;
	const CalendarObject *cb = b;
	time_t diff;
	
	diff = ca->ev_start - cb->ev_start;
	return (diff < 0) ? -1 : (diff > 0) ? 1 : 0;
}

static int
assemble_event_list (MIMEDirVComponent *comp, time_t start, time_t end, void *c)
{
	CalendarObject *co;
	GList **l = c;
	
	co = g_new (CalendarObject, 1);

	co->ev_start = start;
	co->ev_end   = end;
	co->event    = MIMEDIR_VEVENT (comp);
	*l = g_list_insert_sorted (*l, co, calendar_object_compare_by_start);

	return 1;
}

void
calendar_destroy_event_list (GList *l)
{
	GList *p;

	for (p = l; p; p = p->next)
		g_free (p->data);
	g_list_free (l);
}

GList *
calendar_get_events_in_range (Calendar *cal, time_t start, time_t end)
{
	GList *l = 0;
	
	calendar_iterate (cal, start, end, assemble_event_list, &l);
	return l;
}

static GObject *
calendar_object_find_in_list (Calendar *cal, GList *list, const char *euid)
{
	GList *l;

	for (l = list; l; l = l->next){
		GObject *obj;
		gchar *uid;

		g_assert (G_IS_OBJECT (l->data));
		obj = G_OBJECT (l->data);

		g_object_get (obj, "uid", &uid, NULL);

		if (strcmp (uid, euid) == 0) {
			g_free (uid);
			return obj;
		}

		g_free (uid);
	}

	return NULL;
}

GObject *
calendar_object_find_event (Calendar *cal, const char *uid)
{
	g_return_val_if_fail (cal != NULL, NULL);
	g_return_val_if_fail (uid != NULL, NULL);

	return g_hash_table_lookup (cal->event_hash, uid);
}

GObject *
calendar_object_find (Calendar *cal, const char *uid)
{
	GObject *obj;
	
	g_return_val_if_fail (cal != NULL, NULL);
	g_return_val_if_fail (uid != NULL, NULL);

	obj = calendar_object_find_in_list (cal, cal->events, uid);

	return obj;
}
