/* Day view notebook panel for gncal
 *
 * Copyright (C) 1998 The Free Software Foundation
 *
 * Author: Federico Mena <quartic@gimp.org>
 *
 * Changed in 2002 by Sebastian Rittau <srittau@jroger.in-berlin.de>
 * for the GNOME 2 port.
 */

#include <config.h>

#include <time.h>

#include <gnome.h>

#include <mimedir/mimedir-vevent.h>
#include <mimedir/mimedir-vtodo.h>

#include "gncal-day-panel.h"
#include "timeutil.h"
#include "todo-list.h"
#include "utils-time.h"
#include "utils-ui.h"


struct _GncalDayPanelPriv {
	GnomeCalendar *calendar;

	GtkWidget *date_label;
	GtkWidget *fullday_sw;
	GtkWidget *fullday;
	GtkWidget *gtk_calendar;
	GtkWidget *todo;

	/* Signal IDs */

	gulong day_selected_id;
	guint calendar_id;

	time_t start_of_day;
};


static void gncal_day_panel_class_init	(GncalDayPanelClass *klass);
static void gncal_day_panel_init	(GncalDayPanel *dpanel);
static void gncal_day_panel_destroy	(GtkObject *object);
static void gncal_day_panel_finalize	(GObject *object);
static void gncal_day_panel_setup	(GncalDayPanel *dpanel);


static GtkTableClass *parent_class = NULL;


GType
gncal_day_panel_get_type (void)
{
	static GType day_panel_type = 0;

	if (!day_panel_type) {
		static const GTypeInfo day_panel_info = {
			sizeof (GncalDayPanelClass),
			NULL, /* base_init */
			NULL, /* base_finalize */
			(GClassInitFunc) gncal_day_panel_class_init,
			NULL, /* class_finalize */
			NULL, /* class_data */
			sizeof (GncalDayPanel),
			7,    /* n_preallocs */
			(GInstanceInitFunc) gncal_day_panel_init,
		};

		day_panel_type = g_type_register_static (GTK_TYPE_VBOX,
							 "GncalDayPanel",
							 &day_panel_info,
							 0);
	}

	return day_panel_type;
}

static void
gncal_day_panel_class_init (GncalDayPanelClass *klass)
{
	GObjectClass   *gobject_class;
	GtkObjectClass *gtkobject_class;

	g_return_if_fail (klass != NULL);
	g_return_if_fail (GNCAL_IS_DAY_PANEL_CLASS (klass));

	gobject_class   = G_OBJECT_CLASS (klass);
	gtkobject_class = GTK_OBJECT_CLASS (klass);

	gtkobject_class->destroy = gncal_day_panel_destroy;
	gobject_class->finalize = gncal_day_panel_finalize;

	parent_class = g_type_class_peek_parent (klass);
}

static void
gncal_day_panel_init (GncalDayPanel *dpanel)
{
	g_return_if_fail (dpanel != NULL);
	g_return_if_fail (GNCAL_IS_DAY_PANEL (dpanel));

	dpanel->priv = g_new0 (GncalDayPanelPriv, 1);

	gncal_day_panel_setup (dpanel);
}

static void
gncal_day_panel_destroy (GtkObject *object)
{
	GncalDayPanelPriv *priv;

	g_return_if_fail (object != NULL);
	g_return_if_fail (GNCAL_IS_DAY_PANEL (object));

	priv = GNCAL_DAY_PANEL (object)->priv;

	if (priv->gtk_calendar && priv->calendar_id) {
		ui_gconf_shutdown_gtk_calendar (GTK_CALENDAR (priv->gtk_calendar), priv->calendar_id);
		priv->calendar_id = 0;
	}

	GTK_OBJECT_CLASS (parent_class)->destroy (object);
}

static void
gncal_day_panel_finalize (GObject *object)
{
	GncalDayPanel *dpanel;

	g_return_if_fail (object != NULL);
	g_return_if_fail (GNCAL_IS_DAY_PANEL (object));

	dpanel = GNCAL_DAY_PANEL (object);

	g_free (dpanel->priv);

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
day_view_range_activated (GncalFullDay *fullday, GncalDayPanel *dpanel)
{
	MIMEDirVEvent *event;
	MIMEDirDateTime *start, *end;
	time_t dtstart, dtend;

	event = mimedir_vevent_new ();

	gncal_full_day_selection_range (fullday, &dtstart, &dtend);

	start = mimedir_datetime_new_from_time_t (dtstart);
	end   = mimedir_datetime_new_from_time_t (dtend);
	g_object_set (G_OBJECT (event),
		      "dtstart", start,
		      "dtend",   end,
		      NULL);
	g_object_unref (G_OBJECT (start));
	g_object_unref (G_OBJECT (end));

	gnome_calendar_add_object (dpanel->priv->calendar, MIMEDIR_VCOMPONENT (event));
	gncal_full_day_focus_child (fullday, event);
}

static void
full_day_size_allocated (GtkWidget *widget, GtkAllocation *allocation, GncalDayPanel *dpanel)
{
	GtkAdjustment *adj;
	int yoffset;
	gfloat newval;

	adj = gtk_scrolled_window_get_vadjustment (GTK_SCROLLED_WINDOW (dpanel->priv->fullday_sw));

	yoffset = gncal_full_day_get_day_start_yoffset (GNCAL_FULL_DAY (widget));

	newval = adj->lower + (adj->upper - adj->lower) * (double) yoffset / allocation->height;
	if (newval != adj->value)
		g_signal_emit_by_name (G_OBJECT (adj), "value_changed", NULL);
}

static void
calendar_day_selected (GtkCalendar *calendar, GncalDayPanel *dpanel)
{
	gint y, m, d;
	struct tm tm;

	gtk_calendar_get_date (calendar, &y, &m, &d);

	tm.tm_year = y - 1900;
	tm.tm_mon  = m;
	tm.tm_mday = d;
	tm.tm_hour = 5; /* for daylight savings time fix */
	tm.tm_min  = 0;
	tm.tm_sec  = 0;

	gnome_calendar_goto (dpanel->priv->calendar, mktime (&tm));
}

static void
retag_calendar (GtkCalendar *calendar, GncalDayPanel *dpanel)
{
	gnome_calendar_tag_calendar (dpanel->priv->calendar,
				     GTK_CALENDAR (dpanel->priv->gtk_calendar));
}

static void
gncal_day_panel_setup (GncalDayPanel *dpanel)
{
	GncalDayPanelPriv *priv;
	GtkWidget *box;
	GtkWidget *pane;
	gint start_pos = 265;

	priv = dpanel->priv;

	/* General */

	gtk_container_set_border_width (GTK_CONTAINER (dpanel), 12);
	gtk_box_set_spacing (GTK_BOX (dpanel), 12);

	/* Date label */

	priv->date_label = gtk_label_new ("");
	gtk_widget_show (priv->date_label);
	gtk_box_pack_start (GTK_BOX (dpanel), priv->date_label, FALSE, TRUE, 0);

	/* Create horizontal pane */

	pane = gtk_hpaned_new ();
	gtk_paned_set_position (GTK_PANED (pane), start_pos);
	gtk_widget_show (pane);
	gtk_box_pack_start (GTK_BOX (dpanel), pane, TRUE, TRUE, 0);
	
	/* Full day */

	priv->fullday_sw = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (priv->fullday_sw),
					GTK_POLICY_AUTOMATIC,
					GTK_POLICY_AUTOMATIC);
	gtk_widget_show (priv->fullday_sw);
	gtk_paned_pack1 (GTK_PANED (pane), priv->fullday_sw, FALSE, TRUE);

	priv->fullday = gncal_full_day_new ();
	g_signal_connect (G_OBJECT (priv->fullday), "range_activated",
			  G_CALLBACK (day_view_range_activated), dpanel);
	gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (priv->fullday_sw),
					       priv->fullday);
	gtk_widget_show (priv->fullday);

	/* We'll scroll the list to the proper initial position */

	g_signal_connect (G_OBJECT (priv->fullday), "size_allocate",
			  G_CALLBACK (full_day_size_allocated), dpanel);

        /* Create month/todo box */

	box = gtk_vbox_new (FALSE, 12);
	gtk_paned_pack2 (GTK_PANED (pane), box, TRUE, TRUE);

	/* Gtk calendar */

	priv->gtk_calendar = gtk_calendar_new ();
	gtk_calendar_display_options (GTK_CALENDAR (priv->gtk_calendar),
				        GTK_CALENDAR_SHOW_HEADING
				      | GTK_CALENDAR_SHOW_DAY_NAMES);
	priv->day_selected_id =
		g_signal_connect (G_OBJECT (priv->gtk_calendar),
				  "day_selected_double_click",
				  G_CALLBACK (calendar_day_selected), dpanel);
	g_signal_connect (G_OBJECT (priv->gtk_calendar), "month_changed",
			  G_CALLBACK (retag_calendar), dpanel);
	gtk_widget_show (priv->gtk_calendar);
	gtk_box_pack_start (GTK_BOX (box), priv->gtk_calendar, FALSE, FALSE, 0);

	/* To-do */

	priv->todo = todo_list_new ();
	gtk_widget_show (priv->todo);
	gtk_box_pack_start (GTK_BOX (box), priv->todo, TRUE, TRUE, 0);

	/* Configuration */

	priv->calendar_id = ui_gconf_setup_gtk_calendar (GTK_CALENDAR (priv->gtk_calendar));
}

GtkWidget *
gncal_day_panel_new (GnomeCalendar *calendar, time_t start_of_day)
{
	GncalDayPanel *dpanel;

	g_return_val_if_fail (calendar != NULL, NULL);
	g_return_val_if_fail (GNOME_IS_CALENDAR (calendar), NULL);

	dpanel = g_object_new (GNCAL_TYPE_DAY_PANEL, NULL);

	dpanel->priv->calendar = calendar;

	/* FIXME: yikes! This is ugly, ugly, ugly. Fix GncalFullDay to not
	 * require the calendar in this way.
	 */
	GNCAL_FULL_DAY (dpanel->priv->fullday)->calendar = calendar;

	gncal_day_panel_set (dpanel, start_of_day);

	/* FIXME: full day should know about its bounds itself (from gconf) */
	gncal_full_day_set_bounds (GNCAL_FULL_DAY (dpanel->priv->fullday),
				   time_day_begin (start_of_day),
				   time_day_end (start_of_day));

	return GTK_WIDGET (dpanel);
}

static void
gncal_day_panel_update_label (GncalDayPanel *dpanel)
{
	gchar *s, *t;

	s = time_get_date_string (localtime (&dpanel->priv->start_of_day));
	t = g_strconcat ("<b>", s, "</b>", NULL);
	g_free (s);

	gtk_label_set_markup (GTK_LABEL (dpanel->priv->date_label), t);

	g_free (t);
}

void
gncal_day_panel_update (GncalDayPanel *dpanel, MIMEDirVComponent *comp, int flags)
{
	g_return_if_fail (dpanel != NULL);
	g_return_if_fail (GNCAL_IS_DAY_PANEL (dpanel));

	if (MIMEDIR_IS_VEVENT (comp))
		gncal_full_day_update (GNCAL_FULL_DAY (dpanel->priv->fullday),
				       MIMEDIR_VEVENT (comp), flags);
	retag_calendar (GTK_CALENDAR (dpanel->priv->gtk_calendar), dpanel);

	gncal_day_panel_update_label (dpanel);
}

void
gncal_day_panel_set (GncalDayPanel *dpanel, time_t start_of_day)
{
	GncalFullDay *fullday;
	struct tm tm;

	g_return_if_fail (dpanel != NULL);
	g_return_if_fail (GNCAL_IS_DAY_PANEL (dpanel));

	fullday = GNCAL_FULL_DAY (dpanel->priv->fullday);

	dpanel->priv->start_of_day = time_day_begin(start_of_day);
	if (fullday->lower == dpanel->priv->start_of_day)
		return;

	gncal_day_panel_update_label (dpanel);

	/* FIXME: fullday should know about this itself */
	gncal_full_day_set_bounds (fullday,
				   dpanel->priv->start_of_day,
				   time_day_end (dpanel->priv->start_of_day));

	tm = *localtime (&dpanel->priv->start_of_day);

	gtk_calendar_select_month (GTK_CALENDAR (dpanel->priv->gtk_calendar),
				   tm.tm_mon, tm.tm_year + 1900);

	g_signal_handler_block (G_OBJECT (dpanel->priv->gtk_calendar),
				dpanel->priv->day_selected_id);
	gtk_calendar_select_day (GTK_CALENDAR (dpanel->priv->gtk_calendar),
				 tm.tm_mday);
	g_signal_handler_unblock (G_OBJECT (dpanel->priv->gtk_calendar),
				  dpanel->priv->day_selected_id);

	gncal_day_panel_update_label (dpanel);
}

TodoList *
gncal_day_panel_get_todo_list (GncalDayPanel *dpanel)
{
	g_return_val_if_fail (dpanel != NULL, NULL);
	g_return_val_if_fail (GNCAL_IS_DAY_PANEL (dpanel), NULL);

	return TODO_LIST (dpanel->priv->todo);
}
