/* Week view composite widget for gncal
 *
 * Copyright (C) 1998 The Free Software Foundation
 *
 * Authors: Federico Mena <quartic@gimp.org>
 *          Miguel de Icaza <miguel@kernel.org>
 *
 * Changed in 2002 by Sebastian Rittau <srittau@jroger.in-berlin.de>
 * for the GNOME 2 port.
 *
 * $Id: gncal-week-view.c,v 1.31 2002/12/16 13:11:14 srittau Exp $
 */

/* FIXME: honor week start config option */

#include <config.h>

#include <string.h>
#include <time.h>

#include <gnome.h>
#include <gtk/gtk.h>

#include "gncal-week-view.h"
#include "gnomecal-config.h"
#include "timeutil.h"
#include "utils-time.h"
#include "utils-ui.h"


static void gncal_week_view_class_init	(GncalWeekViewClass *klass);
static void gncal_week_view_init	(GncalWeekView *wview);
static void gncal_week_view_destroy	(GtkObject *object);
static void gncal_week_view_finalize	(GObject *object);
static void gncal_week_view_setup	(GncalWeekView *wview);


struct _GncalWeekViewPriv {
	GnomeCalendar *calendar;	/* the calendar we are associated to */

	struct tm start_of_week;

	GtkWidget *days[7];		/* the day view widgets */
	GtkWidget *gtk_calendar;
	GtkWidget *label;

	/* GConf Cache */

	gboolean week_starts_sunday;

	/* Signal IDs */

	guint calendar_id;
	guint notify1;
};


static GtkVBoxClass *parent_class = NULL;


GType
gncal_week_view_get_type (void)
{
	static GType week_view_type = 0;

	if (!week_view_type) {
		static const GTypeInfo week_view_info = {
			sizeof (GncalWeekViewClass),
			NULL, /* base_init */
			NULL, /* base_finalize */
			(GClassInitFunc) gncal_week_view_class_init,
			NULL, /* class_finalize */
			NULL, /* class_data */
			sizeof (GncalWeekView),
			0,    /* n_preallocs */
			(GInstanceInitFunc) gncal_week_view_init,
		};

		week_view_type = g_type_register_static (GTK_TYPE_VBOX,
							 "GncalWeekView",
							 &week_view_info,
							 0);
	}

	return week_view_type;
}

static void
gncal_week_view_class_init (GncalWeekViewClass *klass)
{
	GObjectClass   *gobject_class;
	GtkObjectClass *gtkobject_class;

	g_return_if_fail (klass != NULL);
	g_return_if_fail (GNCAL_IS_WEEK_VIEW_CLASS (klass));

	gobject_class   = G_OBJECT_CLASS   (klass);
	gtkobject_class = GTK_OBJECT_CLASS (klass);

	gobject_class->finalize  = gncal_week_view_finalize;
	gtkobject_class->destroy = gncal_week_view_destroy;

	parent_class = g_type_class_peek_parent (klass);
}

static void
gncal_week_view_init (GncalWeekView *wview)
{
	g_return_if_fail (wview != NULL);
	g_return_if_fail (GNCAL_IS_WEEK_VIEW (wview));

	wview->priv = g_new (GncalWeekViewPriv, 1);

	gncal_week_view_setup (wview);
}

static void
gncal_week_view_destroy (GtkObject *object)
{
	GncalWeekViewPriv *priv;

	g_return_if_fail (object != NULL);
	g_return_if_fail (GNCAL_IS_WEEK_VIEW (object));

	priv = GNCAL_WEEK_VIEW (object)->priv;

	if (priv->gtk_calendar && priv->calendar_id) {
		ui_gconf_shutdown_gtk_calendar (GTK_CALENDAR (priv->gtk_calendar), priv->calendar_id);
		priv->calendar_id = 0;
	}

	GTK_OBJECT_CLASS (parent_class)->destroy (object);
}

static void
gncal_week_view_finalize (GObject *object)
{
	GncalWeekView *wview;
	GConfClient *gconf;

	g_return_if_fail (object != NULL);
	g_return_if_fail (GNCAL_IS_WEEK_VIEW (object));

	wview = GNCAL_WEEK_VIEW (object);

	gconf = gconf_client_get_default ();
	gconf_client_notify_remove (gconf, wview->priv->notify1);
	g_object_unref (G_OBJECT (gconf));

	g_free (wview->priv);

	G_OBJECT_CLASS (parent_class)->finalize (object);
}


static void
gncal_week_view_config_changed (GConfClient *client,
				guint        cnxn_id,
				GConfEntry  *entry,
				gpointer     data)
{
	GncalWeekView *wview;
	const gchar *key;
	gboolean v;

	g_assert (data != NULL);
	g_assert (GNCAL_IS_WEEK_VIEW (data));

	wview = GNCAL_WEEK_VIEW (data);

	key = gconf_entry_get_key (entry);
	v = gconf_value_get_bool (gconf_entry_get_value (entry));

	if (!strcmp (key, CONFIG_GENERAL_WEEK_START))
		wview->priv->week_starts_sunday = v;
}


static void
jump_to_day (GtkCalendar *cal, GncalWeekView *wview, int day)
{
	struct tm tm;
	time_t t;
	
	tm.tm_mday = day;
	tm.tm_mon  = cal->month;
	tm.tm_year = cal->year - 1900;
	tm.tm_hour = 0;
	tm.tm_min  = 0;
	tm.tm_sec  = 0;
	tm.tm_isdst = -1;
	t = mktime (&tm);

	gnome_calendar_goto (wview->priv->calendar, t);
}

static void
jump_to_day_click (GtkCalendar *cal, GncalWeekView *wview)
{
	jump_to_day (cal, wview, cal->selected_day);
}

static void
sync_week (GtkCalendar *cal, GncalWeekView *wview)
{
	jump_to_day (cal, wview, wview->priv->start_of_week.tm_mday + 7);
	gnome_calendar_tag_calendar (wview->priv->calendar,
				     GTK_CALENDAR (wview->priv->gtk_calendar));
}

static void
gncal_week_view_setup (GncalWeekView *wview)
{
	GncalWeekViewPriv *priv;
	GtkWidget *table;
	GConfClient *gconf;
	gint i;

	priv = wview->priv;

	gtk_box_set_spacing (GTK_BOX (wview), 12);

	/* Header label */

	priv->label = gtk_label_new ("");
	gtk_label_set_justify (GTK_LABEL (priv->label), GTK_JUSTIFY_CENTER);
	gtk_widget_show (priv->label);
	gtk_box_pack_start (GTK_BOX (wview), priv->label, FALSE, FALSE, 0);

	/* Table */

	table = gtk_table_new (0, 0, TRUE);
	gtk_widget_show (table);
	gtk_box_pack_start (GTK_BOX (wview), table, TRUE, TRUE, 0);

	/* Days */

	for (i = 0; i < 7; i++) {
		priv->days[i] = gncal_day_view_new ();
		gtk_widget_show (priv->days[i]);

		if (i < 5)
			gtk_table_attach (GTK_TABLE (table), priv->days[i],
					  i, i + 1,
					  0, 1,
					  GTK_EXPAND | GTK_FILL | GTK_SHRINK,
					  GTK_EXPAND | GTK_FILL | GTK_SHRINK,
					  0, 0);
		else
			gtk_table_attach (GTK_TABLE (table), priv->days[i],
					  i - 2, i - 1,
					  1, 2,
					  GTK_EXPAND | GTK_FILL | GTK_SHRINK,
					  GTK_EXPAND | GTK_FILL | GTK_SHRINK,
					  0, 0);
	}

	/* GtkCalendar */

	priv->gtk_calendar = gtk_calendar_new ();

	g_signal_connect (G_OBJECT (priv->gtk_calendar),
			  "day_selected_double_click",
			  G_CALLBACK (jump_to_day_click), wview);
	g_signal_connect (G_OBJECT (priv->gtk_calendar), "month_changed",
			  G_CALLBACK (sync_week), wview);

	gtk_calendar_display_options (GTK_CALENDAR (priv->gtk_calendar),
				        GTK_CALENDAR_SHOW_HEADING
				      | GTK_CALENDAR_SHOW_DAY_NAMES
				      | GTK_CALENDAR_SHOW_WEEK_NUMBERS);

	gtk_widget_show (priv->gtk_calendar);

	gtk_table_attach (GTK_TABLE (table), priv->gtk_calendar,
			  0, 3,
			  1, 2,
			  GTK_EXPAND | GTK_FILL | GTK_SHRINK,
			  GTK_EXPAND | GTK_FILL | GTK_SHRINK,
			  4, 4);

	/* Configuration */

	priv->calendar_id =
		ui_gconf_setup_gtk_calendar (GTK_CALENDAR (priv->gtk_calendar));

	gconf = gconf_client_get_default ();
	priv->week_starts_sunday = gconf_client_get_bool (gconf, CONFIG_GENERAL_WEEK_START, NULL);
	wview->priv->notify1 =
		gconf_client_notify_add (gconf, CONFIG_GENERAL_WEEK_START,
					 gncal_week_view_config_changed, wview,
					 NULL, NULL);
	g_object_unref (G_OBJECT (gconf));
}

GtkWidget *
gncal_week_view_new (GnomeCalendar *calendar, time_t start_of_week)
{
	GncalWeekView *wview;
	int i;

	g_return_val_if_fail (calendar != NULL, NULL);
	g_return_val_if_fail (GNOME_IS_CALENDAR (calendar), NULL);

	wview = g_object_new (GNCAL_TYPE_WEEK_VIEW, NULL);

	wview->priv->calendar = calendar;

	/* FIXME: yada yada: see comment in gncal_day_panel_new() */
	for (i = 0; i < 7; i++) {
		GncalDayView *dview;
		dview = GNCAL_DAY_VIEW (wview->priv->days[i]);
		dview->calendar = calendar;
		gncal_day_view_update (dview, NULL, 0);
	}


	gncal_week_view_set (wview, start_of_week);

	return GTK_WIDGET (wview);
}

void
gncal_week_view_update (GncalWeekView *wview, MIMEDirVComponent *comp, int flags)
{
	int i;

	g_return_if_fail (wview != NULL);
	g_return_if_fail (GNCAL_IS_WEEK_VIEW (wview));

	for (i = 0; i < 7; i++)
		gncal_day_view_update (GNCAL_DAY_VIEW (wview->priv->days[i]),
				       comp, flags);
}

void
gncal_week_view_set (GncalWeekView *wview, time_t start_of_week)
{
	struct tm tm;
	time_t day_start, day_end, week_start, week_end;
	int i;

	g_return_if_fail (wview != NULL);
	g_return_if_fail (GNCAL_IS_WEEK_VIEW (wview));

	localtime_r (&start_of_week, &tm);
	
	/* back up to start of week */

	if (!wview->priv->week_starts_sunday)
		tm.tm_mday -= (tm.tm_wday == 0) ? 6 : (tm.tm_wday - 1);
	else
		tm.tm_mday -= tm.tm_wday;

	/* Start of day */

	tm.tm_hour = 0;
	tm.tm_min  = 0;
	tm.tm_sec  = 0;

	day_start = week_start = mktime (&tm);

	wview->priv->start_of_week = tm;

	/* Calendar */

	gtk_calendar_select_month (GTK_CALENDAR (wview->priv->gtk_calendar),
				   tm.tm_mon, tm.tm_year + 1900);

	/* Day views */

	for (i = 0; i < 7; i++) { /* rest of days */
		tm.tm_mday++;
		day_end = mktime (&tm);

		gncal_day_view_set_bounds (GNCAL_DAY_VIEW (wview->priv->days[i]),
					   day_start, day_end - 1);

		day_start = day_end;
	}
	
	/* The label */
	{
		gchar buffer[20];
		gchar *s1, *s2, *s3, *s4;
		
		week_end = time_add_day (week_start, 6);

		s1 = time_get_date_string (localtime (&week_start));
		s2 = time_get_date_string (localtime (&week_end));

		strftime (buffer, sizeof (buffer), "%W", localtime(&week_start));
		s3 = g_locale_to_utf8 (buffer, -1, NULL, NULL, NULL);

		s4 = g_strdup_printf (_("<b>Week %s</b>\n%s - %s"), s3, s1, s2);

		gtk_label_set_markup (GTK_LABEL (wview->priv->label), s4);

		g_free (s1);
		g_free (s2);
		g_free (s3);
		g_free (s4);
	}
}
