/*
 * GnomeCalendar widget
 * Copyright (C) 1998 the Free Software Foundation
 *
 * Author: Miguel de Icaza (miguel@kernel.org)
 *
 * Changed in 2002 and 2003 by Sebastian Rittau <srittau@jroger.in-berlin.de>
 * for the GNOME 2 port.
 */
#include <config.h>

#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <time.h>

#include <gconf/gconf.h>
#include <gconf/gconf-client.h>
#include <gnome.h>

#include "calendar.h"
#include "gnome-cal.h"
#include "gncal-day-panel.h"
#include "gncal-week-view.h"
#include "gnomecal-alarm-sink.h"
#include "gnomecal-config.h"
#include "gnomecal-ui.h"
#include "month-view.h"
#include "timeutil.h"
#include "utils-time.h"
#include "year-view.h"


static void gnome_calendar_init (GnomeCalendar *gcal);


GType
gnome_calendar_get_type (void)
{
	static GType gnome_calendar_type = 0;
	if(!gnome_calendar_type) {
		GTypeInfo gnome_calendar_info = {
			sizeof(GnomeCalendarClass),
			NULL, /* base_init */
			NULL, /* base_finalize */
			(GClassInitFunc) NULL,
			NULL, /* class_finalize */
			NULL, /* class_data */
			sizeof (GnomeCalendar),
			1,    /* n_preallocs */
			(GInstanceInitFunc) gnome_calendar_init,
		};

		gnome_calendar_type = g_type_register_static (GTK_TYPE_NOTEBOOK,
							      "GnomeCalendar",
							      &gnome_calendar_info,
							      0);
	}
	return gnome_calendar_type;
}

static void
setup_widgets (GnomeCalendar *gcal)
{
	time_t now;

	now = time (NULL);

	gcal->day_view   = gncal_day_panel_new (gcal, now);
	gcal->week_view  = gncal_week_view_new (gcal, now);
	gcal->month_view = month_view_new (gcal, now);
	gcal->year_view  = year_view_new (gcal, now);

	gtk_container_set_border_width (GTK_CONTAINER (gcal->day_view), 12);
	gtk_container_set_border_width (GTK_CONTAINER (gcal->week_view), 12);
	gtk_container_set_border_width (GTK_CONTAINER (gcal->month_view), 12);
	gtk_container_set_border_width (GTK_CONTAINER (gcal->year_view), 12);

	gcal->year_view_sw = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (gcal->year_view_sw),
					GTK_POLICY_NEVER,
					GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER (gcal->year_view_sw), gcal->year_view);
	GTK_LAYOUT (gcal->year_view)->vadjustment->step_increment = 10.0;
	gtk_adjustment_changed (GTK_ADJUSTMENT (GTK_LAYOUT (gcal->year_view)->vadjustment));

	gtk_notebook_append_page (GTK_NOTEBOOK (gcal), gcal->day_view,   gtk_label_new (_("Day View")));
	gtk_notebook_append_page (GTK_NOTEBOOK (gcal), gcal->week_view,  gtk_label_new (_("Week View")));
	gtk_notebook_append_page (GTK_NOTEBOOK (gcal), gcal->month_view, gtk_label_new (_("Month View")));
	gtk_notebook_append_page (GTK_NOTEBOOK (gcal), gcal->year_view_sw,  gtk_label_new (_("Year View")));

	gtk_widget_show_all (GTK_WIDGET (gcal));
}

static void
gnome_calendar_init (GnomeCalendar *gcal)
{
	g_return_if_fail (gcal != NULL);
	g_return_if_fail (GNOME_IS_CALENDAR (gcal));

	gcal->current_day_display = gcal->current_week_display = gcal->current_month_display = 
		gcal->current_year_display = time_day_begin (time (NULL));
	gcal->cal = calendar_new ();
	setup_widgets (gcal);
}

static GtkWidget *
get_current_page (GnomeCalendar *gcal)
{
	gint page_num;

	page_num = gtk_notebook_get_current_page (GTK_NOTEBOOK (gcal));

	return gtk_notebook_get_nth_page (GTK_NOTEBOOK (gcal), page_num);
}

char *
gnome_calendar_get_current_view_name (GnomeCalendar *gcal)
{
	GtkWidget *page;

	g_return_val_if_fail (gcal != NULL, "dayview");
	g_return_val_if_fail (GNOME_IS_CALENDAR (gcal), "dayview");

	page = get_current_page (gcal);
	if (!page)
		return "dayview";

	if (page == gcal->day_view)
		return "dayview";
	else if (page == gcal->week_view)
		return "weekview";
	else if (page == gcal->month_view)
		return "monthview";
	else if (page == gcal->year_view_sw)
		return "yearview";
	else
		return "dayview";
}

void
gnome_calendar_goto (GnomeCalendar *gcal, time_t new_time)
{
	GtkWidget *current;

	g_return_if_fail (gcal != NULL);
	g_return_if_fail (GNOME_IS_CALENDAR (gcal));
	g_return_if_fail (new_time != -1);

	current = get_current_page (gcal);
	if (!current)
		return;
	
	new_time = time_day_begin (new_time);

	if (current == gcal->day_view) {
		gncal_day_panel_set (GNCAL_DAY_PANEL (gcal->day_view), new_time);
		gcal->current_day_display = new_time;
	} else if (current == gcal->week_view) {
		gncal_week_view_set (GNCAL_WEEK_VIEW (gcal->week_view), new_time);
		gcal->current_week_display = new_time;
	} else if (current == gcal->month_view) {
		month_view_set (MONTH_VIEW (gcal->month_view), new_time);
		gcal->current_month_display = new_time;
	} else if (current == gcal->year_view_sw) {
		year_view_set (YEAR_VIEW (gcal->year_view), new_time);
		gcal->current_year_display = new_time;
	} else
		g_return_if_reached ();
}

static void
gnome_calendar_go_one_step (GnomeCalendar *gcal, gboolean step_back)
{
	GtkWidget *cp;
	time_t new_time;
	gint step;

	cp = get_current_page (gcal);
	if (!cp)
		return;

	if (step_back)
		step = -1;
	else
		step = 1;

	if (cp == gcal->day_view)
		new_time = time_add_day (gcal->current_day_display, step);
	else if (cp == gcal->week_view)
		new_time = time_add_week (gcal->current_week_display, step);
	else if (cp == gcal->month_view)
		new_time = time_add_month (gcal->current_month_display, step);
	else if (cp == gcal->year_view_sw)
		new_time = time_add_year (gcal->current_year_display, step);
	else
		g_return_if_reached ();

	gnome_calendar_goto (gcal, new_time);
}

void
gnome_calendar_next (GnomeCalendar *gcal)
{
	g_return_if_fail (gcal != NULL);
	g_return_if_fail (GNOME_IS_CALENDAR (gcal));

	gnome_calendar_go_one_step (gcal, FALSE);
}

void
gnome_calendar_previous (GnomeCalendar *gcal)
{
	g_return_if_fail (gcal != NULL);
	g_return_if_fail (GNOME_IS_CALENDAR (gcal));

	gnome_calendar_go_one_step (gcal, TRUE);
}

void
gnome_calendar_dayjump (GnomeCalendar *gcal, time_t time)
{
	g_return_if_fail (gcal != NULL);
	g_return_if_fail (GNOME_IS_CALENDAR (gcal));

	gtk_notebook_set_current_page (GTK_NOTEBOOK (gcal), 0);
	gnome_calendar_goto (gcal, time);
}

void
gnome_calendar_goto_today (GnomeCalendar *gcal)
{
	g_return_if_fail (gcal != NULL);
	g_return_if_fail (GNOME_IS_CALENDAR (gcal));

	gnome_calendar_goto (gcal, time (NULL));
}

void
gnome_calendar_set_view (GnomeCalendar *gcal, const gchar *page_name)
{
	int page = 0;

	g_return_if_fail (gcal != NULL);
	g_return_if_fail (GNOME_IS_CALENDAR (gcal));
	g_return_if_fail (page_name != NULL);


	if (strcmp (page_name, "dayview") == 0)
		page = 0;
	else if (strcmp (page_name, "weekview") == 0)
		page = 1;
	else if (strcmp (page_name, "monthview") == 0)
		page = 2;
	else if (strcmp (page_name, "yearview") == 0)
		page = 3;
	gtk_notebook_set_current_page (GTK_NOTEBOOK (gcal), page);
}

GtkWidget *
gnome_calendar_new (void)
{
	GnomeCalendar  *gcal;

	gcal = g_object_new (GNOME_TYPE_CALENDAR, NULL);

	return GTK_WIDGET (gcal);
}

void
gnome_calendar_update_all (GnomeCalendar *cal, MIMEDirVComponent *component, int flags)
{
	gncal_day_panel_update (GNCAL_DAY_PANEL (cal->day_view), component, flags);
	gncal_week_view_update (GNCAL_WEEK_VIEW (cal->week_view), component, flags);
	month_view_update (MONTH_VIEW (cal->month_view), component, flags);
	year_view_update (YEAR_VIEW (cal->year_view), component, flags);
}

void
gnome_calendar_add_object (GnomeCalendar *gcal, MIMEDirVComponent *comp)
{
	g_return_if_fail (gcal != NULL);
	g_return_if_fail (GNOME_IS_CALENDAR (gcal));
	g_return_if_fail (comp != NULL);
	g_return_if_fail (MIMEDIR_IS_VCOMPONENT (comp));

	calendar_add_object (gcal->cal, comp);
	gnome_calendar_update_all (gcal, comp, CHANGE_NEW);
}

void
gnome_calendar_remove_object (GnomeCalendar *gcal, MIMEDirVComponent *comp)
{
	g_return_if_fail (gcal != NULL);
	g_return_if_fail (GNOME_IS_CALENDAR (gcal));
	g_return_if_fail (comp != NULL);
	g_return_if_fail (MIMEDIR_IS_VCOMPONENT (comp));

	calendar_remove_object (gcal->cal, comp);
	gnome_calendar_update_all (gcal, comp, CHANGE_ALL);
}

void
gnome_calendar_object_changed (GnomeCalendar *gcal, MIMEDirVComponent *comp, int flags)
{
	g_return_if_fail (gcal != NULL);
	g_return_if_fail (GNOME_IS_CALENDAR (gcal));
	g_return_if_fail (comp != NULL);
	g_return_if_fail (MIMEDIR_IS_VCOMPONENT (comp));

	gcal->cal->modified = TRUE;

	gnome_calendar_update_all (gcal, comp, flags);
}

/*
 * called from the calendar_iterate routine to mark the days of a GtkCalendar
 */
static int
mark_gtk_calendar_day (MIMEDirVComponent *component, time_t start, time_t end, void *c)
{
	GtkCalendar *gtk_cal = c;
	struct tm tm_s;
	time_t t, day_end;

	tm_s = *localtime (&start);
	day_end = time_day_end (end);

	for (t = start; t < day_end; t += 60*60*24){
		time_t new = mktime (&tm_s);
		struct tm tm_day;

		tm_day = *localtime (&new);
		gtk_calendar_mark_day (gtk_cal, tm_day.tm_mday);
		tm_s.tm_mday++;
	}
	return TRUE;
}

/*
 * Tags the dates with appointments in a GtkCalendar based on the
 * GnomeCalendar contents
 */
void
gnome_calendar_tag_calendar (GnomeCalendar *cal, GtkCalendar *gtk_cal)
{
	time_t month_begin, month_end;
	struct tm tm;

	g_return_if_fail (cal != NULL);
	g_return_if_fail (GNOME_IS_CALENDAR (cal));
	g_return_if_fail (gtk_cal != NULL);
	g_return_if_fail (GTK_IS_CALENDAR (gtk_cal));

	/* compute month_begin */
	tm.tm_hour = 0;
	tm.tm_min  = 0;
	tm.tm_sec  = 0;
	tm.tm_mday = 1; /* setting this to zero is a no-no; it will set mktime back to the end of the
					   previous month, which may be 28,29,30; this may chop some days from the calendar */
	tm.tm_mon  = gtk_cal->month;
	tm.tm_year = gtk_cal->year - 1900;
	tm.tm_isdst= -1;

	month_begin = mktime (&tm);
	tm.tm_mon++;
	month_end   = mktime (&tm);

	gtk_calendar_freeze (gtk_cal);
	gtk_calendar_clear_marks (gtk_cal);
	calendar_iterate (cal->cal, month_begin, month_end, mark_gtk_calendar_day, gtk_cal);
	gtk_calendar_thaw (gtk_cal);
}

time_t
gnome_calendar_current_display (GnomeCalendar *gcal)
{
	GtkWidget *page;

	g_return_val_if_fail (gcal != NULL, -1);
	g_return_val_if_fail (GNOME_IS_CALENDAR (gcal), -1);

	page = get_current_page (gcal);
	if (!page)
		return -1;

	if (page == gcal->day_view)
		return gcal->current_day_display;
	if (page == gcal->week_view)
		return gcal->current_week_display;
	if (page == gcal->month_view)
		return gcal->current_month_display;
	if (page == gcal->year_view_sw)
		return gcal->current_year_display;

	g_assert_not_reached();
	return -1;
}
	
