/* Miscellaneous utility functions for the calendar view widgets
 *
 * Copyright (C) 1998 The Free Software Foundation
 *
 * Authors: Federico Mena <quartic@gimp.org>
 *          Miguel de Icaza <miguel@kernel.org>
 *
 * Changed in 2002 by Sebastian Rittau <srittau@jroger.in-berlin.de>
 * for the GNOME 2 port.
 */

#include <string.h>

#include <gnome.h>

#include <mimedir/mimedir-vevent.h>

#include "utils-time.h"
#include "view-utils.h"


typedef struct {
	PangoLayout *layout;
	int lines;
	int assigned_lines;
} line_info_t;

void
view_utils_draw_events (GtkWidget *widget, GdkWindow *window, GdkGC *gc, GdkRectangle *area,
			int flags, GList *events, time_t start, time_t end)
{
	GtkStyle *style;
	gint fontsize;
	int y, max_y, items, i, need_more, nlines, base, extra;
	GList *list;
	line_info_t *lines;

	if (events == NULL)
		return;
	
	style = gtk_widget_get_style (widget);

	items = g_list_length (events);
	lines = g_new0 (line_info_t, items);

	fontsize = pango_font_description_get_size (style->font_desc) / PANGO_SCALE + 1;

	max_y = area->y + area->height - fontsize;
	if (flags & VIEW_UTILS_DRAW_SPLIT)
	    max_y *= 2;

	/*
	 * Layout all the lines, measure the space needs
	 */
	for (i = 0, list = events; list; list = list->next, i++){
		CalendarObject *co = list->data;
		MIMEDirVEvent *event = co->event;
		GString *buf;
		gchar *s, *summary;
		struct tm tm_start, tm_end;

		buf = g_string_new (NULL);

		tm_start = *localtime (&co->ev_start);
		tm_end   = *localtime (&co->ev_end);

		s = time_get_short_time_string (&tm_start);
		g_string_append (buf, s);
		g_free (s);
			
		if (flags & VIEW_UTILS_DRAW_END){
			g_string_append_c (buf, '-');
			s = time_get_short_time_string (&tm_end);
			g_string_append (buf, s);
			g_free (s);
		}

		g_object_get (G_OBJECT (event), "summary", &summary, NULL);
		if (mimedir_vcomponent_get_allday (MIMEDIR_VCOMPONENT (event))) {
			g_string_assign (buf, summary);
		} else {
			g_string_append (buf, ": ");
			g_string_append (buf, summary);
		}
		g_free (summary);

		lines[i].layout = gtk_widget_create_pango_layout (widget,
								  buf->str);
		pango_layout_set_width  (lines[i].layout, area->width * PANGO_SCALE);
		pango_layout_set_wrap   (lines[i].layout, PANGO_WRAP_WORD);
		pango_layout_set_indent (lines[i].layout, -4 * PANGO_SCALE);
		lines[i].lines = pango_layout_get_line_count (lines[i].layout);

		g_string_free (buf, TRUE);
	}

	/*
	 * Compute how many lines we will give to each row
	 */
	nlines = 1 + max_y / fontsize;
	base = nlines / items;
	extra = nlines % items;
	need_more = 0;

	for (i = 0; i < items; i++){
		if (lines [i].lines <= base){
			extra += base - lines [i].lines;
			lines [i].assigned_lines = lines [i].lines;
		} else {
			need_more++;
			lines [i].assigned_lines = base;
		}
	}

	/*
	 * use any extra space
	 */
	while (need_more && extra > 0){
		need_more = 0;
		
		for (i = 0; i < items; i++){
			if (lines [i].lines > lines [i].assigned_lines){
				lines [i].assigned_lines++;
				extra--;
			}

			if (extra == 0)
				break;
			
			if (lines [i].lines > lines [i].assigned_lines)
				need_more = 1;
		}
	}

	/*
	 * Draw the information
	 */
	gdk_gc_set_clip_rectangle (gc, area);
	y = area->y + fontsize;
	for (i = 0; i < items; i++){
		int line;
		GSList *line_list;

		line_list = pango_layout_get_lines (lines[i].layout);

		for (line = 0; line < lines[i].assigned_lines; line++) {
			gdk_draw_layout_line (window, gc,
					  area->x, y,
					  line_list->data);
			y += fontsize;
			line_list = line_list->next;
		}
	}

	gdk_gc_set_clip_rectangle (gc, NULL);

	/*
	 * Free resources.
	 */
	g_free (lines);	
}

void
view_utils_draw_textured_frame (GtkWidget *widget, GdkWindow *window, GdkRectangle *rect, GtkShadowType shadow)
{
	int x, y;
	int xthick, ythick;
	GdkGC *light_gc, *dark_gc;
	GtkStyle *style;

	style = gtk_widget_get_style (widget);

	gdk_draw_rectangle (window,
			    style->bg_gc[GTK_STATE_NORMAL],
			    TRUE,
			    rect->x, rect->y,
			    rect->width, rect->height);

	light_gc = style->light_gc[GTK_STATE_NORMAL];
	dark_gc  = style->dark_gc[GTK_STATE_NORMAL];

	xthick = style->xthickness;
	ythick = style->ythickness;

	gdk_gc_set_clip_rectangle (light_gc, rect);
	gdk_gc_set_clip_rectangle (dark_gc, rect);

	for (y = rect->y + ythick; y < (rect->y + rect->height - ythick); y += 3)
		for (x = rect->x + xthick; x < (rect->x + rect->width - xthick); x += 6) {
			gdk_draw_point (window, light_gc, x, y);
			gdk_draw_point (window, dark_gc, x + 1, y + 1);

			gdk_draw_point (window, light_gc, x + 3, y + 1);
			gdk_draw_point (window, dark_gc, x + 4, y + 2);
		}
	
	gdk_gc_set_clip_rectangle (light_gc, NULL);
	gdk_gc_set_clip_rectangle (dark_gc, NULL);

	gtk_paint_shadow (widget->style, window,
			  GTK_STATE_NORMAL, shadow,
			  NULL, NULL, NULL,
			  rect->x, rect->y,
			  rect->width, rect->height);
}
